from unittest import TestCase

from taco import Taco
from taco.object import TacoObject

class TacoClientActionTestCase(TestCase):
    def test_call_class_method(self):
        t = DummyClient()
        t.call_class_method('tc', 'tm', 1, 2, 3, four=4, five=5)
        self.assertEqual(t.msg, {
            'action': 'call_class_method',
            'class': 'tc',
            'name': 'tm',
            'args': (1, 2, 3),
            'kwargs': {'four': 4, 'five': 5},
            'context': None,
        })

    def test_call_function(self):
        t = DummyClient()
        t.call_function('tf', 1, 2, 3, four=4, five=5)
        self.assertEqual(t.msg, {
            'action': 'call_function',
            'name': 'tf',
            'args': (1, 2, 3),
            'kwargs': {'four': 4, 'five': 5},
            'context': None,
        })

    def test_call_method(self):
        t = DummyClient()
        o = TacoObject(t, 4444)
        o.call_method('tm', 1, 2, 3, four=4, five=5)
        self.assertEqual(t.msg, {
            'action': 'call_method',
            'number': 4444,
            'name': 'tm',
            'args': (1, 2, 3),
            'kwargs': {'four': 4, 'five': 5},
            'context': None,
        })

    def test_construct_object(self):
        t = DummyClient()
        t.construct_object('tc', 5, 6, 7, 8, x=111, y=222)
        self.assertEqual(t.msg, {
            'action': 'construct_object',
            'class': 'tc',
            'args': (5, 6, 7, 8),
            'kwargs': {'x': 111, 'y': 222},
        })

    def test_destroy_object(self):
        t = DummyClient()
        o = TacoObject(t, 55555)
        del o
        self.assertEqual(t.msg, {
            'action': 'destroy_object',
            'number': 55555,
        })

    def test_get_attribute(self):
        t = DummyClient()
        o = TacoObject(t, 666666)
        o.get_attribute('ta')
        self.assertEqual(t.msg, {
            'action': 'get_attribute',
            'number': 666666,
            'name': 'ta',
        })

    def test_get_value(self):
        t = DummyClient()
        t.get_value('tv')
        self.assertEqual(t.msg, {
            'action': 'get_value',
            'name': 'tv',
        })

    def test_import_module(self):
        t = DummyClient()
        t.import_module('module.name')
        self.assertEqual(t.msg, {
            'action': 'import_module',
            'name': 'module.name',
            'args': (),
            'kwargs': {},
        })

    def test_set_attrbute(self):
        t = DummyClient()
        o = TacoObject(t, 7777777)
        o.set_attribute('ta', 88)
        self.assertEqual(t.msg, {
            'action': 'set_attribute',
            'number': 7777777,
            'name': 'ta',
            'value': 88,
        })

    def test_set_value(self):
        t = DummyClient()
        t.set_value('value.name', 999)
        self.assertEqual(t.msg, {
            'action': 'set_value',
            'name': 'value.name',
            'value': 999,
        })

class TacoClientConvenienceTestCase(TestCase):
    def test_function(self):
        t = DummyClient()
        f = t.function('convfunc')
        f('arg1', 'arg2', kwarg1=9, kwarg2=99)
        self.assertEqual(t.msg, {
            'action': 'call_function',
            'name': 'convfunc',
            'args': ('arg1', 'arg2'),
            'kwargs': {'kwarg1': 9, 'kwarg2': 99},
            'context': None,
        })

    def test_constructor(self):
        t = DummyClient()
        c = t.constructor('convclass')
        c(3, 3, 3, x=1, y=2)
        self.assertEqual(t.msg, {
            'action': 'construct_object',
            'class': 'convclass',
            'args': (3, 3, 3),
            'kwargs': {'x': 1, 'y': 2},
        })

class DummyClient(Taco):
    def __init__(self):
        self.msg = None

    def _interact(self, msg):
        self.msg = msg



# Taco Python client module.
# Copyright (C) 2014 Graham Bell
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class TacoError(Exception):
    """Base class for specific Taco client exceptions.

    Note that the client can also raise general exceptions,
    such as ValueError, if its methods are called with invalid
    arguments.
    """

    pass


class TacoReceivedError(TacoError):
    """An exception raised by the Taco client.

    Raised if the client receives an exception action.  The exception
    message will contain any message text received in the exception action.
    """

    pass


class TacoUnknownActionError(TacoError):
    """An exception raised by the Taco client.

    Raised if the client receives an action of an unknown type.
    """

    pass



from unittest import TestCase
try:
    from unittest.mock import Mock
except ImportError:
    from mock import Mock

from taco.object import TacoObject

class TacoObjectTestCase(TestCase):
    def test_create_destroy(self):
        client = Mock()

        obj = TacoObject(client, 77)

        self.assertIsInstance(obj, TacoObject)

        del obj

        client._destroy_object.assert_called_with(77)

    def test_methods(self):
        client = Mock()

        obj = TacoObject(client, 84)

        obj.call_method('test_method', 1, 2, three=3, four=4)
        client._call_method.assert_called_with(
            84, 'test_method', 1, 2, three=3, four=4)

        obj.set_attribute('test_attribute', 4444)
        client._set_attribute.assert_called_with(
            84, 'test_attribute', 4444)

        g_a = Mock(return_value=55555)
        client._get_attribute = g_a
        self.assertEqual(obj.get_attribute('test_attribute2'), 55555)
        g_a.assert_called_with(84, 'test_attribute2')

    def test_convenience(self):
        client = Mock()

        obj = TacoObject(client, 333)

        conv = obj.method('conv')

        conv(5, 6, 7, 8)

        client._call_method.assert_called_with(333, 'conv', 5, 6, 7, 8)



from collections import namedtuple
from datetime import date, datetime
from io import BytesIO
from unittest import TestCase
from sys import version, version_info

from taco.server import TacoServer
from taco.object import TacoObject

from . import DummyBase

class TacoServerMethodTestCase(TestCase):
    def test_server_construction(self):
        ts = DummyServer()

        self.assertEqual(ts._null_result,
            {'action': 'result', 'result': None})

    def test_find_attr(self):
        ts = DummyServer()

        self.assertEqual(ts._find_attr('__name__'), 'taco.server')

        self.assertIs(ts._find_attr('zip'), zip)

        ts.ns['test_var'] = 5678

        self.assertEqual(ts._find_attr('test_var'), 5678)

        self.assertEqual(ts._find_attr('sys.version_info.minor'),
                         version_info.minor)

    def test_object_handling(self):
        ts = DummyServer()

        self.assertEqual(ts.nobject, 0)

        d = date(2000, 4, 1)
        ts.xp.write({'object': d})

        self.assertEqual(ts.nobject, 1)

        self.assertEqual(ts.get_output(),
                        '{"object": {"_Taco_Object_": 1}}')

        ts.prepare_input('{"object": {"_Taco_Object_": 1}}')
        r = ts.xp.read()

        self.assertIs(r['object'], d)

class TacoServerActionTestCase(TestCase):
    def test_call_class_method(self):
        ts = DummyServer()
        ts.import_module({
            'name': 'datetime',
            'args': ['datetime'],
        })
        self.assertEqual(ts.call_class_method({
            'class': 'datetime',
            'name': 'strptime',
            'args': ['2011-07-31', '%Y-%m-%d'],
            'kwargs': {},
        })['result'], datetime(2011, 7, 31))

    def test_call_function(self):
        ts = DummyServer()
        self.assertEqual(ts.call_function({
            'name': 'divmod',
            'args': [67, 8],
            'kwargs': {},
        })['result'], (8, 3))

    def test_call_method(self):
        ts = DummyServer()
        ts.objects[1] = datetime(2020, 2, 20)
        self.assertEqual(ts.call_method({
            'number': 1,
            'name': 'strftime',
            'args': ['%m/%d/%Y'],
            'kwargs': {},
        })['result'], '02/20/2020')

    def test_construct_object(self):
        ts = DummyServer()
        ts.import_module({
            'name': 'datetime',
            'args': ['datetime'],
        })
        self.assertEqual(ts.construct_object({
            'class': 'datetime',
            'args': [1999, 12, 31],
            'kwargs': {'second': 59},
        })['result'], datetime(1999, 12, 31, second=59))

    def test_destroy_object(self):
        ts = DummyServer()
        ts.objects[1] = [1]
        self.assertEqual(ts.destroy_object({'number': 1})['result'], None)
        self.assertNotIn(1, ts.objects)

    def test_get_attribute(self):
        ts = DummyServer()
        TT = namedtuple('TT', ['xyz'])
        ts.objects[777] = TT(888)
        self.assertEqual(ts.get_attribute({
            'number': 777,
            'name': 'xyz',
        })['result'], 888)

    def test_get_value(self):
        ts = DummyServer()
        ts.import_module({'name': 'sys', 'args': [], 'kwargs': {}})
        self.assertEqual(ts.get_value({
            'name': 'sys.version'
        })['result'], version)

    def test_import_module(self):
        ts = DummyServer()
        self.assertNotIn('datetime', ts.ns)
        ts.import_module({'name': 'datetime', 'args': []})
        self.assertIn('datetime', ts.ns)
        self.assertNotIn('date', ts.ns)
        ts.import_module({'name': 'datetime', 'args': ['date']})
        self.assertIn('date', ts.ns)

    def test_set_attribute(self):
        ts = DummyServer()
        ob = NumberObject(44)
        ts.objects[9] = ob
        ts.set_attribute({
            'number': 9,
            'name': 'number',
            'value': 55,
        })
        self.assertEqual(ob.number, 55)

    def test_set_value(self):
        ts = DummyServer()
        d = {'xxx': 'yyy'}
        ts.ns['zzz'] = d
        ts.set_value({
            'name': 'zzz.ppp',
            'value': 'qqq'
        })
        self.assertEqual(d['ppp'], 'qqq')

class DummyServer(TacoServer, DummyBase):
    def _construct_transport(self):
        self.in_ = BytesIO()
        self.out = BytesIO()

        return TacoServer._construct_transport(self, self.in_, self.out)

class NumberObject():
    def __init__(self, number):
        self.number = number



from codecs import utf_8_decode, utf_8_encode

class DummyBase():
    def prepare_input(self, string):
        self.in_.seek(0)
        self.in_.write(utf_8_encode(string + '\n// END\n')[0])
        self.in_.truncate()
        self.in_.seek(0)

    def get_output(self):
        r = utf_8_decode(self.out.getvalue())[0]
        self.out.seek(0)
        self.out.truncate()
        if not r.endswith('\n// END\n'):
            raise Exception('end marker not found')
        return r[:-8]



# Taco Python object module.
# Copyright (C) 2013 Graham Bell
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class TacoObject():
    """Taco object class.

    This class is used to represent objects by Taco actions.
    Instances of this class will returned by methods of Taco
    objects and should not normally be constructed explicitly.

    The objects reside on the server side and are referred to by instances
    of this class by their object number.  When these instances are
    destroyed the destroy_object action is sent automatically.
    """

    def __init__(self, client, number):
        """Construct new reference object.

        Stores a reference to the Taco client to allow actions to be
        sent via it.
        """

        self.client = client
        self.number = number

    def __del__(self):
        """Destroy the object by passing the destroy_object message to the
        server."""

        self.client._destroy_object(self.number)

    def call_method(self, *args, **kwargs):
        """Invoke the given method on the object.

        The first argument is the method name.

        The context (void / scalar / list) can be specified as a
        keyword argument "context" unless the "disable_context" attribute
        of the client has been set.
        """

        return self.client._call_method(self.number, *args, **kwargs)

    def get_attribute(self, *args, **kwargs):
        """Retrieve the value of the given attribute."""

        return self.client._get_attribute(self.number, *args, **kwargs)

    def set_attribute(self, *args, **kwargs):
        """Set the value of the given attribute."""

        return self.client._set_attribute(self.number, *args, **kwargs)

    def method(self, name):
        """Convenience method giving a function which calls a method.

        Returns a function which can be used to invoke a method on
        the server object. For example::

            strftime = afd.method('strftime')
            print(strftime('%Y-%m-%d'))
        """

        def func(*args, **kwargs):
            return self.call_method(name, *args, **kwargs)

        return func



# Taco Python transport module.
# Copyright (C) 2013 Graham Bell
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from codecs import utf_8_decode, utf_8_encode
from json import JSONDecoder, JSONEncoder


class TacoTransport():
    """Taco transport class.

    Implements the communication between Taco clients and servers.
    """

    def __init__(self, in_, out, from_obj=None, to_obj=None):
        """Constructs new TacoTransport object.

        Stores the input and output streams.  The "from_obj" and "to_obj"
        functions are used for the JSON encoder's "default" function
        and the JSON decoder's "object_hook" parameters respectively.
        """

        self.in_ = in_
        self.out = out

        self.encoder = JSONEncoder(default=from_obj)
        self.decoder = JSONDecoder(object_hook=to_obj)

    def read(self):
        """Read a message from the input stream.

        The decoded message is returned as a data structure, or
        None is returned if nothing was read.
        """

        text = ''
        while True:
            line = self.in_.readline()
            line = utf_8_decode(line)[0]

            if line == '' or line.startswith('// END'):
                break

            text += line

        if text == '':
            return None

        return self.decoder.decode(text)

    def write(self, message):
        """Write a message to the output stream."""

        message = self.encoder.encode(message)
        self.out.write(utf_8_encode(message)[0])
        self.out.write(utf_8_encode('\n// END\n')[0])
        self.out.flush()



# Taco Python server module.
# Copyright (C) 2013-2014 Graham Bell
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

try:
    import builtins
except ImportError:
    import __builtin__ as builtins

import sys

from taco.transport import TacoTransport


class TacoServer():
    """Taco server class.

    This class implements a Taco server for Python.
    """

    def __init__(self):
        """Construct TacoServer object.

        This method also calls the _construct_transport method
        to construct a transport object.
        """

        self.ns = {}
        self.objects = {}
        self.nobject = 0

        self._null_result = self._make_result(None)

        self.xp = self._construct_transport()

    def _construct_transport(self, in_=None, out=None):
        """Create TacoTransport object.

        Constructs a TacoTransport object communicating via standard
        input and standard output.

        sys.stdout is set to sys.stderr to try to avoid text being
        written to standard output, which would corrupt communications
        with the client.
        """

        if in_ is None or out is None:
            # When running under Python 3, detach the text wrappers from
            # stdin and stdout so that we have binary streams which act
            # like other streams which TacoTransport needs to handle.
            try:
                in_ = sys.stdin.detach()
                out = sys.stdout.detach()
            except AttributeError:
                in_ = sys.stdin
                out = sys.stdout

            # Redirect stdout to stderr to prevent any called functions from
            # writing into the Taco message stream.
            sys.stdout = sys.stderr

        def from_obj(obj):
            """Place object in the objects dictionary and return a Taco object
            reference."""

            self.nobject += 1
            self.objects[self.nobject] = obj
            return {'_Taco_Object_': self.nobject}

        def to_obj(dict_):
            """If dict_ is a Taco object reference, fetch the corresponding
            object from objects dictionary."""

            if '_Taco_Object_' in dict_:
                return self.objects[dict_['_Taco_Object_']]
            else:
                return dict_

        return TacoTransport(in_, out, from_obj, to_obj)

    def run(self):
        """Main server function.

        Enters a message handling loop.  The loop exits on failure to
        read another message.
        """

        while True:
            message = self.xp.read()

            if message is None:
                break

            act = message['action']

            if hasattr(self, act) and not act.startswith('_'):
                try:
                    res = getattr(self, act)(message)
                except Exception as e:
                    res = {
                        'action': 'exception',
                        'message': 'exception caught: ' + str(e),
                    }
            else:
                res = {
                    'action': 'exception',
                    'message': 'unknown action: ' + act,
                }

            self.xp.write(res)

    def _make_result(self, result):
        """Construct Taco result message."""

        return {
            'action': 'result',
            'result': result,
        }

    def _find_attr(self, name):
        """Attempt to look up the given name.

        The name is split into parts on the "." character and the root part
        is searched for in the "ns" dictionary, in globals() and finally in
        builtins.  Once the root part is found, attributes corresponding to the
        remainder of the name are looked up.  An exception is raised
        (explicity) if the root part cannot be found or (implicitly)
        if one of the remaining parts cannot be found.
        """

        parts = name.split('.')
        root = parts.pop(0)

        if root in self.ns:
            result = self.ns[root]
        elif root in globals():
            result = globals()[root]
        elif hasattr(builtins, root):
            result = getattr(builtins, root)
        else:
            raise Exception('cannot find "{0}"'.format(root))

        for part in parts:
            result = getattr(result, part)

        return result

    def call_class_method(self, message):
        """Call the class method specified in the message.

        The context, if present in the message, is ignored.
        """

        cls = self._find_attr(message['class'])
        func = getattr(cls, message['name'])
        return self._make_result(func(
            *(message['args'] if message['args'] is not None else ()),
            **(message['kwargs'] if message['kwargs'] is not None else {})))

    def call_function(self, message):
        """Call the function specified in the message.

        The context, if present in the message, is ignored.
        """

        func = self._find_attr(message['name'])
        return self._make_result(func(
            *(message['args'] if message['args'] is not None else ()),
            **(message['kwargs'] if message['kwargs'] is not None else {})))

    def call_method(self, message):
        """Call an object method.

        Works similarly to call_function.
        """

        obj = self.objects[message['number']]
        return self._make_result(getattr(obj, message['name'])(
            *(message['args'] if message['args'] is not None else ()),
            **(message['kwargs'] if message['kwargs'] is not None else {})))

    def construct_object(self, message):
        """Call an object constructor.

        Works similarly to call_function.
        """

        cls = self._find_attr(message['class'])
        return self._make_result(cls(
            *(message['args'] if message['args'] is not None else ()),
            **(message['kwargs'] if message['kwargs'] is not None else {})))

    def destroy_object(self, message):
        """Remove an object from the objects dictionary."""

        del self.objects[message['number']]
        return self._null_result

    def get_attribute(self, message):
        """Get an attribute value from an object."""

        return self._make_result(getattr(self.objects[message['number']],
                                         message['name']))

    def get_value(self, message):
        """Get the value of a variable.

        If the variable name contains "."-separated components, then it is
        looked up using the _find_attr function.
        """

        (root, _, name) = message['name'].rpartition('.')

        if root:
            base = self._find_attr(root)
        else:
            base = self.ns

        try:
            return self._make_result(base[name])
        except TypeError:
            return self._make_result(getattr(base, name))

    def import_module(self, message):
        """Import a module or names from a module.

        Without arguments, the module is imported and the top level package
        name is inserted into the "ns" dictionary.

        With "args" specified, it is used as a list of names to import
        from the module, and those names are inserted into the "ns"
        dictionary.

        Currently any "kwargs" in the message are ignored.
        """

        if message['args']:
            # With arguments, emulate: from name import args

            mod = __import__(message['name'], fromlist=message['args'],
                             level=0)
            for name in message['args']:
                self.ns[name] = getattr(mod, name)
        else:
            # Without arguments do plain import
            mod = __import__(message['name'], level=0)
            self.ns[mod.__name__] = mod

        return self._null_result

    def set_attribute(self, message):
        """Set an attribute value of an object."""

        setattr(self.objects[message['number']],
                message['name'], message['value'])
        return self._null_result

    def set_value(self, message):
        """Set the value of a variable.

        If the variable name contains "."-separated components, then it is
        looked up using the _find_attr function.
        """

        (root, _, name) = message['name'].rpartition('.')

        if root:
            base = self._find_attr(root)
        else:
            base = self.ns

        try:
            base[name] = message['value']
        except TypeError:
            setattr(base, name, message['value'])

        return self._null_result



# Taco Python client module.
# Copyright (C) 2013-2014 Graham Bell
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess

from taco.error import TacoReceivedError, TacoUnknownActionError
from taco.object import TacoObject
from taco.transport import TacoTransport


class Taco():
    """Taco client class.

    Example::

        from taco import Taco

        taco = Taco(lang='python')

        taco.import_module('time', 'sleep')
        taco.call_function('sleep', 5)
    """

    def __init__(self, lang=None, script=None, disable_context=False):
        """Construct new Taco client by connecting to a server instance.

        The server script can either be specified explicitly with the
        "script" argument, or the language can be given with the "lang"
        argument.  In that case the server script will be assumed to
        be named taco-language and install in your binary search path
        ($PATH). If neither of these arguments is given, then a
        ValueError will be raised.

        The server script will be launched using subprocess.Popen
        and a TacoTransport object will be connected to it.
        """

        if script is not None:
            scr = script
        elif lang is not None:
            scr = 'taco-' + lang
        else:
            raise ValueError('language or script not specified')

        self.disable_context = disable_context

        p = subprocess.Popen([scr],
                             stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE)

        self.xp = self._construct_transport(p.stdout, p.stdin)

    def _construct_transport(self, in_, out):
        """Prepare a TacoTransport object for use with this client.

        from_obj and to_obj functions are passed to the TacoTransport
        constructor to allow the JSON encoder/decoder to deal with Taco
        object references.
        """

        def from_obj(obj):
            if isinstance(obj, TacoObject):
                return {'_Taco_Object_': obj.number}
            else:
                raise ValueError('can not send non-Taco object')

        def to_obj(dict_):
            if '_Taco_Object_' in dict_:
                return TacoObject(self, dict_['_Taco_Object_'])
            else:
                return dict_

        return TacoTransport(in_, out, from_obj, to_obj)

    def _interact(self, message):
        """Private general interaction method used to implement other methods.

        Writes the given message to the server and reads the response.
        If this is a result, then its value is returned.  If it is an
        exception, then a TacoError exception is raised.
        """

        self.xp.write(message)

        response = self.xp.read()

        action = response['action']

        if action == 'result':
            return response['result']
        elif action == 'exception':
            raise TacoReceivedError('received exception: ' +
                                    response['message'])
        else:
            raise TacoUnknownActionError('received unknown action: ' + action)

    def call_class_method(self, class_, name, *args, **kwargs):
        """Invoke a class method call in the connected server.

        The context (void / scalar / list) can be specified as a
        keyword argument "context" unless the "disable_context" attribute
        has been set.
        """

        if 'context' in kwargs and not self.disable_context:
            context = kwargs.pop('context')
        else:
            context = None

        return self._interact({
            'action': 'call_class_method',
            'class': class_,
            'name': name,
            'args': args,
            'kwargs': kwargs,
            'context': context,
        })

    def call_function(self, name, *args, **kwargs):
        """Invoke a function call in the connected server.

        The context (void / scalar / list) can be specified as a
        keyword argument "context" unless the "disable_context" attribute
        has been set.
        """

        if 'context' in kwargs and not self.disable_context:
            context = kwargs.pop('context')
        else:
            context = None

        return self._interact({
            'action': 'call_function',
            'name': name,
            'args': args,
            'kwargs': kwargs,
            'context': context,
        })

    def _call_method(self, number, name, *args, **kwargs):
        """Private method for TacoObjects to send the call_method action."""

        if 'context' in kwargs and not self.disable_context:
            context = kwargs.pop('context')
        else:
            context = None

        return self._interact({
            'action': 'call_method',
            'number': number,
            'name': name,
            'args': args,
            'kwargs': kwargs,
            'context': context,
        })

    def construct_object(self, class_, *args, **kwargs):
        """Invoke an object constructor.

        If successful, this should return a :class:`~taco.object.TacoObject`
        instance which references the new object.  The given arguments and
        keyword arguments are passed to the object constructor.
        """

        return self._interact({
            'action': 'construct_object',
            'class': class_,
            'args': args,
            'kwargs': kwargs,
        })

    def _destroy_object(self, number):
        """Private method for TacoObjects to send the destroy_object action."""
        self._interact({
            'action': 'destroy_object',
            'number': number,
        })

    def _get_attribute(self, number, name):
        """Private method for TacoObjects to send the get_attribute action."""
        return self._interact({
            'action': 'get_attribute',
            'number': number,
            'name': name,
        })

    def get_value(self, name):
        """Request the value of the given variable."""

        return self._interact({
            'action': 'get_value',
            'name': name,
        })

    def import_module(self, name, *args, **kwargs):
        """Instruct the server to load the specified module.

        The interpretation of the arguments depends on the language of
        the Taco server implementation.
        """

        self._interact({
            'action': 'import_module',
            'name': name,
            'args': args,
            'kwargs': kwargs,
        })

    def _set_attribute(self, number, name, value):
        """Private method for TacoObjects to send the set_attribute action."""
        self._interact({
            'action': 'set_attribute',
            'number': number,
            'name': name,
            'value': value,
        })

    def set_value(self, name, value):
        """Set the value of the given variable."""

        self._interact({
            'action': 'set_value',
            'name': name,
            'value': value,
        })

    def function(self, name):
        """Convience method giving a function which calls call_function.

        This example is equivalent to that given for this class::

            sleep = taco.function('sleep')
            sleep(5)
        """

        def func(*args, **kwargs):
            return self.call_function(name, *args, **kwargs)

        return func

    def constructor(self, class_):
        """Convience method giving a function which calls construct_object.

        For example constructing multiple datetime objects::

            taco.import_module('datetime', 'datetime')
            afd = taco.construct_object('datetime', 2000, 4, 1)

        Could be done more easily::

            datetime = taco.constructor('datetime')
            afd = datetime(2000, 4, 1)
        """

        def func(*args, **kwargs):
            return self.construct_object(class_, *args, **kwargs)

        return func



from codecs import utf_8_decode, utf_8_encode
from io import BytesIO
from unittest import TestCase

from taco.transport import TacoTransport

class TacoTransportTestCase(TestCase):
    def test_transport(self):
        in_ = BytesIO(utf_8_encode('{"test_input":1}\n// END\n')[0])
        out = BytesIO()

        xp = TacoTransport(in_, out)

        xp.write({'test_output': 2})

        r = utf_8_decode(out.getvalue())[0]

        self.assertEqual(r, "{\"test_output\": 2}\n// END\n")

        r = xp.read()

        self.assertEqual(r, {'test_input': 1})



from io import BytesIO
from unittest import TestCase

from taco import Taco
from taco.object import TacoObject

from . import DummyBase

# This assert method was renamed in Python 3.2.
if not hasattr(TestCase, 'assertRaisesRegex'):
    TestCase.assertRaisesRegex = TestCase.assertRaisesRegexp

class TacoClientMethodTestCase(TestCase):
    def test_interaction(self):
        t = DummyClient()

        t.prepare_input('{"action": "result", "result": 46}')

        self.assertEqual(t._interact({'action': 'test'}), 46)
        self.assertEqual(t.get_output(), '{"action": "test"}')

        t.prepare_input('{"action": "non-existent action"}')

        with self.assertRaisesRegex(Exception, 'received unknown action'):
            t._interact({'action': 'test'})

        t.prepare_input('{"action": "exception", "message": "test_exc"}')

        with self.assertRaisesRegex(Exception, 'test_exc'):
            t._interact({'action': 'test'})

    def test_object_handing(self):
        t = DummyClient()

        t.prepare_input('{"action": "result", "result": null}')
        o = TacoObject(t, 345)
        t._interact({'test_object': o})
        self.assertEqual(t.get_output(),
            '{"test_object": {"_Taco_Object_": 345}}')

        t.prepare_input('{"action": "result", "result": {"_Taco_Object_": 78}}')
        r = t._interact({'action': 'test'})

        self.assertIsInstance(r, TacoObject)
        self.assertEqual(r.number, 78)

class DummyClient(Taco, DummyBase):
    def __init__(self):
        self.disable_context = False
        self.in_ = BytesIO()
        self.out = BytesIO()

        self.xp = self._construct_transport(self.in_, self.out)

    def _destroy_object(self, number):
        pass



#!/usr/bin/env python

from distutils.core import setup

with open('README.rst') as f:
    long_description = f.read()

setup(
    name='taco',
    version='0.1.0',
    description='Taco module for Python',
    long_description=long_description,
    author='Graham Bell',
    author_email='graham.s.bell@gmail.com',
    url='http://github.com/grahambell/taco-python',
    package_dir={'': 'lib'},
    packages=['taco'],
    scripts=['scripts/taco-python'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
